(* 11:37 Dec 16, 2012 hxy, wzh *)

open Type
open Sast

type symbol_table = {
	parent : symbol_table option;
	variables : (t * string) list;	(* variable type * variable name *)
	in_loop : bool;			(* true if in a loop *)
	in_switch : bool			(* true if in a switch *)
}

type function_table = {			(* reture type * function name * formals *)
	functions : (t * string * t list) list;
}

let string_of_type = function
  Void_T -> "void"
| Int_T -> "int"
| Char_T -> "char"
| Float_T -> "float"
| String_T -> "string"
| Boolean_T -> "boolean"
| Point_T -> "point"
| Color_T -> "color"
| Key_T -> "key"

let string_of_unop = function
  Neg -> "\'-\'"
| Not -> "\'!\'"
| Atx -> "\'@x\'"
| Aty -> "\'@y\'"
| Atr -> "\'@r\'"
| Atg -> "\'@g\'"
| Atb -> "\'@b\'"
| Atk -> "\'@k\'"

let string_of_binop = function
  Add -> "\'+\'"
| Sub -> "\'-\'"
| Mul -> "\'*\'"
| Div -> "\'/\'"
| Mod -> "\'%\'"
| Gt -> "\'>\'"
| Lt -> "\'<\'"
| Geq -> "\'>=\'"
| Leq -> "\'<=\'"
| Eq -> "\'==\'"
| Neq -> "\'!=\'"
| And -> "\'&&\'"
| Or -> "\'||\'"

let oprand_check_unop t unop =	(* unop : Type.unop -> unit = <fun>, check operand type of unary operator *)
	let correct =
		match unop with
		  Neg -> (match t with Int_T | Float_T -> true | _ -> false)
		| Not -> (t = Boolean_T)
		| Atx | Aty -> (t = Point_T)
		| Atr | Atg | Atb -> (t = Color_T)
		| Atk -> (t = Key_T) in
	if not correct then
		raise (Failure("The oprand of " ^ string_of_unop unop ^ " can not be \'" ^ string_of_type t ^ "\'"))

let oprand_check_binop t leftside binop =	(* binop : Type.binop -> unit = <fun>, check operand type of binary operator *)
	let correct = 
		match binop with		(* '+' for string is to concatenate two strings *)
		  Add -> (match t with Int_T | Char_T | Float_T | Point_T | Color_T | String_T -> true | _ -> false) 
		| Sub -> (match t with Int_T | Char_T | Float_T | Point_T | Color_T -> true | _ -> false)
		| Mul | Div -> (match t with Int_T | Float_T -> true | _ -> false)
		| Mod -> (t = Int_T)
		| Gt | Lt | Geq | Leq -> (match t with Int_T | Char_T | Float_T | String_T -> true | _ -> false)
		| Eq | Neq -> (match t with 
			Int_T | Char_T | Float_T | String_T | Boolean_T | Point_T | Color_T | Key_T -> true | _ -> false)
		| And | Or -> (t = Boolean_T) in
	if not correct then
		if leftside = true then
			raise (Failure("The left oprand of " ^ string_of_binop binop ^ " can not be \'" ^ string_of_type t ^ "\'"))
		else raise (Failure("The right oprand of " ^ string_of_binop binop ^ " can not be \'" ^ string_of_type t ^ "\'"))

let rec find_variable scope name = 
	try
		List.find (fun (_, s) -> s = name) scope.variables
	with Not_found ->
		match scope.parent with
		  Some(parent) -> find_variable parent name
		| _ -> raise Not_found

let rec expr scope funcs = function	(* AST.expr -> SAST.expr *)
  Ast.Id(id) ->
	(try
		let (var_type, _) = find_variable scope id in
			(Sast.Id(id), var_type, false)
	with Not_found ->
		raise (Failure ("\'" ^ id ^ "\' was not declared in this scope")))
| Ast.Literal(lit) ->
	let lit_type = 
		match lit with
		  Integer(i) -> Int_T
		| Char(c) -> Char_T
		| Float(f) -> Float_T
		| String(str) -> String_T
		| Boolean(b) -> Boolean_T
		| Point(x,y) -> Point_T
		| Color(r,g,b) -> Color_T
		| Key(key) -> Key_T
	in (Sast.Literal(lit), lit_type, true)
| Ast.Assign(vname, e2) ->
	let new_e1 = expr scope funcs (Ast.Id(vname)) and new_e2 = expr scope funcs e2 in
		let (_, t1, _) = new_e1 and (_, t2, _) = new_e2 in
			if t1 = t2 then 
				(Sast.Assign(vname, new_e2), t1, false)
			else
				raise(Failure ("type mismatch of \'" ^ string_of_type t1 
					^ "\' and \'" ^ string_of_type t2 ^ "\' in the assignment of \'" ^ vname ^ "\'"))
| Ast.Binop(e1, binop, e2) ->
	let new_e1 = expr scope funcs e1 and new_e2 = expr scope funcs e2 in
		let (_, t1, c1) = new_e1 and (_, t2, c2) = new_e2 in
			let _ = oprand_check_binop t1 true binop and _ = oprand_check_binop t2 false binop in (* check left & right oprands *)
				if t1 = t2 then
					(Sast.Binop(new_e1, binop, new_e2), (fun t ->
						match binop with
						  Gt | Lt | Geq | Leq | Eq | Neq -> Boolean_T
						| _ -> t) t1, c1 && c2) (* constant expr if both oprands are constant expr *)
				else
					raise (Failure("type mismatch of \'" ^ string_of_type t1 
					^ "\' and \'" ^ string_of_type t2 ^ "\' for the binary operator " ^ string_of_binop binop))
| Ast.Unop(unop, e) ->
	let new_e = expr scope funcs e in
		let (_, t, c) = new_e in
			let _ = oprand_check_unop t unop in
				(Sast.Unop(unop, new_e), (fun t ->
						match unop with
						  Neg | Not -> t
						| Atx | Aty | Atr | Atg | Atb -> Int_T
						| Atk -> String_T) t, c)
| Ast.Call(fname, actuals) ->
	let (return_type, _, formals_types) = try
		List.find (fun (_, s, _) -> s = fname) funcs.functions
	with Not_found ->
		raise(Failure("\'" ^ fname ^ "\' was not declared in this scope")) in 
	let new_actuals = List.map (expr scope funcs) actuals in
		let actuals_types = List.map (fun (_, t, _) -> t) new_actuals in
			if formals_types = actuals_types then
				(Sast.Call(fname, new_actuals), return_type, false)
			else
				raise(Failure("type mismatch between formals and actuals of function \'" ^ fname ^ "\'"))
| Ast.NoExpr -> (Sast.NoExpr, Void_T, false)

let init_transform scope funcs = function	(* Ast.expr -> Sast.expr for initialization of variable decleration *)
  Some(e) -> Some(expr scope funcs e)
| None -> None

let rec count_var_break scope num =
	if scope.in_loop || scope.in_switch then 
		num
	else
		match scope.parent with
		  Some(parent)-> count_var_break parent (num + List.length scope.variables)
		| None -> raise(Failure("break statement not within loop or switch"))

let rec count_var_continue scope num =
	if scope.in_loop then 
		num
	else
		match scope.parent with
		  Some(parent)-> count_var_continue parent (num + List.length scope.variables)
		| None -> raise(Failure("continue statement not within a loop"))

let var_decl_transform decl =	(* Ast.var_decl -> Sast.var_decl, needs init_transform to transform init *)
	let t = fst decl in
		List.rev (List.fold_left (fun l -> function
		  (name, Ast.NoExpr) -> (t, name, None)::l
		| (name, e) -> (t, name, Some(e))::l) [] (snd decl))
		
let add_var (scope, funcs, new_var_list) (t, name, e) = 
	try 
		let _ =
			List.find (fun (_, n) -> n = name) scope.variables
		in raise(Failure("redeclaration of \'" ^ string_of_type t ^ " " ^ name ^ "\'"))
	with Not_found -> 
		let new_scope = { scope with 
			variables = (t, name)::scope.variables
		} in 
		let new_init = init_transform scope funcs e in
		let init_t =
		match new_init with
		  Some(a, b, c) -> b 
		| None -> t in (* No need to check types, thus t = init_t is trivial *)
		let new_var_decl = 
			if t = init_t then
				(t, name, new_init)
			else
				raise(Failure ("type mismatch of \'" ^ string_of_type t 
					^ "\' and \'" ^ string_of_type init_t ^ "\' in the assignment of \'" ^ name ^ "\'"))
			in (new_scope, funcs, new_var_decl::new_var_list)

let get_inherit_list = function
  [] -> []
| l -> List.map (fun (t, s, _) -> (t, s)) l

let rec stmt scope funcs inherit_list = function	(* AST.stmt -> SAST.stmt *)
  Ast.Expr(e) -> Sast.Expr(expr scope funcs e)
| Ast.Block(block) -> 
	let block_scope = { parent = Some(scope) ; variables = get_inherit_list inherit_list ; in_loop = false; in_switch = false } in
		let check_stmt (scope', st1) s = 
			match s with
			  Ast.VarDecl(var_decl) ->	(* variable declarations can update the scope *)
				let var_list = var_decl_transform var_decl in
					let (new_scope, _, new_var_list) = List.fold_left add_var (scope', funcs, []) var_list
						in (new_scope, (List.map (fun x -> Sast.VarDecl(x)) new_var_list) @ st1) 
			| _ -> (scope', (stmt scope' funcs [] s)::st1) in
			let (new_block_scope, st1) = List.fold_left check_stmt (block_scope, []) block
				in Sast.Block(List.rev st1, List.rev new_block_scope.variables)
| Ast.Return(e) -> Sast.Return(expr scope funcs e)
| Ast.If(exprlist, stmtlist) ->
	let exprList1 = (List.rev (List.fold_left (fun l e ->
		let expr1 = expr scope funcs e in
			let (_, t, _) = expr1 in
				if t = Boolean_T then 
					expr1::l 
				else raise(Failure("predicates of if statments must be type \'boolean\'"))) [] exprlist)) in
		let stmtlist1 = List.rev (List.fold_left (fun l s->
			let st1 = stmt scope funcs [] s in
				let stmt1 =
					match st1 with
					  Sast.Block(_, _) -> st1
					| _ -> raise(Failure("unexpected error"))
				in stmt1::l) [] stmtlist)
					in Sast.If(exprList1, stmtlist1)  
| Ast.For(expr1, expr2, expr3, stmt1) ->
	let e1 = expr scope funcs expr1 in
		let e2 = expr scope funcs expr2 in
			let (_, t2, _) = e2 in
				if t2 = Boolean_T then
					let for_scope = { scope with in_loop = true }	in
					let block_for =
						match stmt1 with
						  Ast.Block(_) -> stmt1
						| _ -> raise(Failure("unexpected error")) in
					let e3 = expr scope funcs expr3 in
						let new_stmt = stmt for_scope funcs [] block_for in
							Sast.For (e1, e2, e3, new_stmt)
				else
					raise(Failure("the second expression in for statements must be type \'boolean\'"))
| Ast.ForWithDecl(var_decl, expr1, expr2, stmt1) ->
	let var_list = var_decl_transform var_decl in
		let (new_scope, _, new_var_list) = List.fold_left add_var (scope, funcs, []) var_list in
			let e1 = expr new_scope funcs expr1 in
				let (_, t1, _) = e1 in
					if t1 = Boolean_T then
						let for_scope = { scope with in_loop = true } in (* original scope as the parent *)
						let block_for =
							match stmt1 with
							  Ast. Block(_) -> stmt1
							| _ -> raise(Failure("unexpected error")) in
						let e2 = expr new_scope funcs expr2 in
						let new_stmt = stmt for_scope funcs new_var_list block_for in
							Sast.ForWithDecl(List.rev new_var_list, e1, e2, new_stmt)
					else
						raise(Failure("the second expression in for statements must be type \'boolean\'"))
| Ast.While(expr1, stmt1) ->
	let new_expr = expr scope funcs expr1 in
		let (_, t1, _) = new_expr in
			if t1 = Boolean_T then
				let loop_scope = { scope with in_loop = true }
					in let block_while = 
					match stmt1 with
					  Ast.Block(_) -> stmt1
					| _ -> raise(Failure("unexpected error"))
				in let new_stmt = stmt loop_scope funcs [] block_while
					in Sast.While (new_expr, new_stmt)
			else
				raise (Failure("predicates of while statments must be type \'boolean\'"))
| Ast.Switch(expr1, exprlist, stmtlist) ->
	let e' = expr scope funcs expr1 in
		let (_, t', _ ) = e' in
			if t' != Float_T then  
				let exprList1 = (List.rev (List.fold_left (fun l e ->
				let e1 = expr scope funcs e in
				let (_, t1, c1) = e1 in
				if c1 = false then
					raise(Failure("case label must be a constant-expression"))
				else if t1 != t' then
					raise(Failure("case label must have the same type as switch quantity"))
				else
					e1::l) [] exprlist)) in
				let new_scope = { scope with in_switch = true } in
				let stmtlist1 = List.map (fun case_stmt ->
					let case_block =
						match case_stmt with
						  Ast.Block(_) -> case_stmt
						| _ -> raise(Failure("unexpected error"))
					in stmt new_scope funcs [] case_block
				) stmtlist 
					in Sast.Switch(e', exprList1, stmtlist1)
			else raise(Failure("switch quantity can not be a float"))
| Ast.Continue -> Sast.Continue(count_var_continue scope 0)
| Ast.Break -> Sast.Break(count_var_break scope 0)
| Ast.VarDecl(_) -> raise(Failure("unexpected error"))		(* Ast.VarDecl is handled in Ast.Block *)

let rec check_return fdecl = function 		(* check whether return statement exists *)
  Sast.Expr(e) -> false
| Sast.Block(s_list, _) -> List.fold_left (fun a b -> a || b) false (List.map (check_return fdecl) s_list)
| Sast.Return(e) -> let (_, t, _) = e in
	if t = fdecl.Ast.rtype then true else raise(Failure ("return expression has the type \'" ^ string_of_type t ^ 
						"\' but a return expression was expected of type \'" ^ string_of_type fdecl.Ast.rtype ^
						"\' in function \'" ^ string_of_type fdecl.Ast.rtype ^ " " ^
						fdecl.Ast.fname ^ "(" ^ ((fun l -> String.concat ", " 
						(List.map string_of_type (List.map fst l))) fdecl.Ast.formals) ^ ")\'"))
| Sast.If(e_list, s_list) -> List.fold_left (fun a b -> a || b) false (List.map (check_return fdecl) s_list)
| Sast.For(e1, e2, e3, s) -> check_return fdecl s
| Sast.ForWithDecl(vdecl_list, e1, e2, s) -> check_return fdecl s
| Sast.While(e, s) -> check_return fdecl s
| Sast.Switch(e, e_list, s_list) -> List.fold_left (fun a b -> a || b) false (List.map (check_return fdecl) s_list)
| Sast.Continue(i) -> false
| Sast.Break(i) -> false
| Sast.VarDecl(vdecl) -> false

let semantically_check program =		(* AST.program -> SAST.program *)
	let init_funcs = { functions = [
		(Void_T, "__leftClickPoint", [Point_T]);
		(Void_T, "__rightClickPoint", [Point_T]);
		(Void_T, "__doubleClick", [Point_T]);
		(Void_T, "__movePoint", [Point_T]);
		(Void_T, "__leftDown", []);
		(Void_T, "__rightDown", []);
		(Void_T, "__leftUp", []);
		(Void_T, "__rightUp", []);
		(Void_T, "__keyDown", [Key_T]);
		(Void_T, "__keyUp", [Key_T]);
		(Void_T, "__keyStroke", [Key_T]);
		(Void_T, "__comboStroke", [Key_T; Key_T; Key_T; Key_T; Key_T; Key_T; Key_T]);
		(Point_T, "__getPoint", []);
		(Color_T, "__getcolor", [Point_T]);
		(Void_T, "__print", [String_T]);
		(String_T, "__stringOfInt", [Int_T]);
		(String_T, "__stringOfFloat", [Float_T]);
		(String_T, "__stringOfChar", [Char_T]);
		(String_T, "__stringOfBoolean", [Boolean_T]);
		(String_T, "__stringOfPoint", [Point_T]);
		(String_T, "__stringOfKey", [Key_T]);
		(String_T, "__stringOfColor", [Color_T]);
		(Int_T, "__getASCII", [Char_T]);
		(Char_T, "__getChar", [Int_T]);
		(Int_T, "__getStringLength", [String_T]);
		(Char_T, "__getCharOfString", [String_T; Int_T]);
		(String_T, "__setCharOfString", [String_T; Int_T; Char_T]);
		(String_T, "__subOfString", [Int_T; Int_T]);
		(Void_T, "__delay", [Int_T]);
		] } in
	let init_scope = {
		parent = None;
		variables = [];
		in_loop = false;
		in_switch = false } in
	let (new_var_decl, new_func_decl) = 
		let rec check var_decl func_decl scope funcs = function
		  [] -> (var_decl, List.rev func_decl)
		| Ast.Var_decl(vdecl)::tl -> 
			let new_vdecl = var_decl_transform vdecl in
				let sast_vdecl = List.map (fun (t, name, init) -> 
				try
					let _ = List.find (fun (_, s) -> s = name) scope.variables
						in raise(Failure("redeclaration of \'" ^ string_of_type t ^ " " ^ name ^ "\'"))
				with Not_found -> 
					let new_init = init_transform scope funcs init in
					let init_t =
					match new_init with
					  Some(a, b, c) -> b 
					| None -> t in (* No need to check types, thus t = init_t is trivial *)
						if t = init_t then
							(t, name, new_init)
						else
							raise(Failure ("type mismatch of \'" ^ string_of_type t 
								^ "\' and \'" ^ string_of_type init_t ^ "\' in the assignment of \'" ^ name ^ "\'"))
					) new_vdecl in
				let updated_scope = { scope with
					variables = scope.variables @ List.map (fun (t, name, _) -> (t, name)) new_vdecl
				}
				in check (var_decl @ sast_vdecl) func_decl updated_scope funcs tl
		| Ast.Func_decl(fdecl)::tl -> 
			try
				let _ = List.find (fun (_, s, _) -> s = fdecl.Ast.fname) funcs.functions
					in raise(Failure ("redeclaration of function \'" ^ string_of_type fdecl.Ast.rtype ^ " " ^
						fdecl.Ast.fname ^ "(" ^ ((fun l -> String.concat ", " (List.map string_of_type
						(List.map fst l))) fdecl.Ast.formals) ^ ")\'"))
			with Not_found ->
				let rec formals_name_test = function
				  [] -> (false, "")
				| hd::tl -> let name = snd hd in
					if List.exists (fun (_, s) -> s = name) tl then 
						(true, name)
					else formals_name_test tl in
				let (same_name, name) = formals_name_test fdecl.Ast.formals in
					if same_name then 
						raise(Failure("\'" ^ name ^ "\' has a previous declaration"))
					else
				let formals_types = List.map fst fdecl.Ast.formals in
					let updated_funcs = {
						functions = (fdecl.Ast.rtype, fdecl.Ast.fname, formals_types)::funcs.functions
					} in
					let func_scope = { scope with
						parent = Some(scope);
						variables = fdecl.Ast.formals
					} in
					let sast_block = stmt func_scope updated_funcs [] (Ast.Block(fdecl.Ast.body)) in
					let valid_return = check_return fdecl sast_block in
						if(not valid_return) then 
							raise(Failure ("no return statement in function \'" ^ string_of_type fdecl.Ast.rtype ^
							" " ^	fdecl.Ast.fname ^ "(" ^ ((fun l -> String.concat ", " (List.map string_of_type 								(List.map fst l))) fdecl.Ast.formals) ^ ")\'"))
						else
					let (sast_body, sast_locals) = match sast_block with
					  Sast.Block(stmts_list, locals_list) ->
						List.iter (fun (_, name) ->
							try
								let _ = List.find (fun (_, s) -> s = name) locals_list
							in raise(Failure("declaration of \'" ^ name ^ "\' shadows a parameter"))
							with Not_found -> ()) fdecl.Ast.formals; (stmts_list, locals_list)
					| _ -> raise(Failure("unexpected error")) in
					let sast_fdecl = {
						rtype = fdecl.Ast.rtype;
						fname = fdecl.Ast.fname;
						formals = fdecl.Ast.formals;
						locals = sast_locals;
						body = sast_body
					} in check var_decl (sast_fdecl::func_decl) scope updated_funcs tl
		in check [] [] init_scope init_funcs (List.rev program)
	in (new_var_decl, new_func_decl)
